package org.ow2.odis.joram;

import java.net.ConnectException;
import java.net.UnknownHostException;
import java.util.Observable;
import java.util.Observer;

import javax.jms.Connection;
import javax.jms.JMSException;
import javax.jms.Session;
import javax.jms.TopicConnectionFactory;

import org.objectweb.joram.client.jms.ConnectionFactory;
import org.objectweb.joram.client.jms.Destination;
import org.objectweb.joram.client.jms.admin.AdminException;
import org.objectweb.joram.client.jms.admin.AdminModule;
import org.objectweb.joram.client.jms.ha.tcp.TopicHATcpConnectionFactory;
import org.objectweb.util.monolog.Monolog;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.odis.admin.Launch;
import org.ow2.odis.exception.OdisAttributeException;
import org.ow2.odis.exception.OdisConnectionException;
import org.ow2.odis.exception.OdisJmsException;
import org.ow2.odis.lifeCycle.connection.AbstractConnectionState;
import org.ow2.odis.lifeCycle.connection.StateConnectionLoaded;
import org.ow2.odis.model.AbstractConnectionAttribute;


/**
 * The Class AbstractConnexionJMS.
 * @author Beyler Olivier FranceTelecom R&D base class of ConnexionJMS
 */
public abstract class AbstractConnectionJMS implements Observer {
    /** Logger for this class. */
    static final Logger loggerCnx = Monolog.initialize().getLogger(AbstractConnectionJMS.class.getName());

    /** The Constant CST_JORAM_SERV_ERR. */
    private static final String CST_JORAM_SERV_ERR = "Your Joram version seem to be incompatible or the Joram server is down, check your jar lib";

    /** The Constant NOT_READY. */
    public static final int NOT_READY = -1;

    /** The Constant READY. */
    public static final int READY = 0;

    /** State of the connection. */
    private AbstractConnectionState cnxState = new StateConnectionLoaded();

    /** Contains all information about a JMS destination. */
    protected final AbstractJMSConnectionAttribute attribute;

    /** JMS session. */
    protected transient Session session = null;

    /** A destination is the base class of Queue or Topic. */
    protected transient Destination destination = null;

    /** Connection used by JMS to send or get message. */
    protected transient Connection cnx = null;

    /** The state. */
    private int state = READY;

    /**
     * Instantiates a new abstract connection JMS.
     * @param attribute the attribute
     */
    public AbstractConnectionJMS(final AbstractJMSConnectionAttribute attribute) {
        super();
        this.attribute = attribute;
        attribute.addObserver(this);
        if (loggerCnx.isLoggable(BasicLevel.DEBUG)) {
            loggerCnx.log(BasicLevel.DEBUG, "Construct ConnexionJMS");
        }
    }

    /**
     * Connects to the Joram JMS AdminModule. Three cases exist :
     * <li>1) joram is collocated
     * <li>2) joram is not collocated
     * <li>3) joram is a HA cluster.
     * @throws ConnectException the connect exception
     * @throws AdminException the admin exception
     * @throws UnknownHostException the unknown host exception
     */
    protected void connection() throws ConnectException, AdminException, UnknownHostException {
        if (loggerCnx.isLoggable(BasicLevel.INFO)) {
            loggerCnx.log(BasicLevel.INFO, "Connexion on adminModule " + attribute.getProxyServerHostName() + ":"
                    + attribute.getProxyServerPort());
        }
        if ("".equals(attribute.getHAClusterDescription())) {
            if (loggerCnx.isLoggable(BasicLevel.INFO)) {
                loggerCnx.log(BasicLevel.INFO, "Joram is " + (Launch.joramCollocated ? "" : "NON ") + "collocated");
            }
            if (Launch.joramCollocated) {
                AdminModule.collocatedConnect(attribute.getUserName(), attribute.getUserPassword());
            } else {
                AdminModule.connect(attribute.getProxyServerHostName(), attribute.getProxyServerPort(), attribute.getUserName(), attribute
                        .getUserPassword(), 50);
            }
        } else {
            if (loggerCnx.isLoggable(BasicLevel.INFO)) {
                loggerCnx.log(BasicLevel.INFO, "Joram is CLUSTERED");
            }
            final TopicConnectionFactory tcf = TopicHATcpConnectionFactory.create(attribute.getHAClusterDescription());
            ((ConnectionFactory) tcf).getParameters().connectingTimer = 50;
            AdminModule.connect(tcf, attribute.getUserName(), attribute.getUserPassword());
        }
    }

    /**
     * Gets the attribute.
     * @return Returns the attribute.
     */
    public AbstractConnectionAttribute getAttribute() {
        return attribute;
    }

    /**
     * getter.
     * @return Session
     */
    public Session getSession() {
        return session;
    }

    /**
     * create or retrieve the Queue/Topic on the joram server. In case on
     * @throws OdisJmsException if the Joram Server can not be reached
     */
    synchronized void createDestination() throws OdisJmsException {
        try {
            connection();
        } catch (Exception e) {
            loggerCnx.log(BasicLevel.FATAL, CST_JORAM_SERV_ERR, e);
            throw new OdisJmsException(CST_JORAM_SERV_ERR, e);
        }
        if (cnx == null) {
            destination = null;
            if (loggerCnx.isLoggable(BasicLevel.INFO)) {
                loggerCnx.log(BasicLevel.INFO, "AbstractConnectionJMS trying to connect : " + attribute.getName());
            }
            cnx = ConnectionJMSManager.getInstance().connect(this);
        }
        if (cnx == null) {
            throw new OdisJmsException("Couldn't obtain a connection (cnx == null)");
        }
        try {
            if (loggerCnx.isLoggable(BasicLevel.INFO)) {
                loggerCnx.log(BasicLevel.INFO, "Using connection : " + cnx);
            }
            specificCreateDestination();
            if (loggerCnx.isLoggable(BasicLevel.INFO)) {
                loggerCnx.log(BasicLevel.INFO, "After JMS specificCreateDestination() : " + attribute.getName() + "/" + cnx + "/"
                        + session);
            }
        } catch (ConnectException e) {
            StringBuffer buff = new StringBuffer("ConnectException [");
            buff.append(attribute.getName());
            buff.append("] unable to discuss with admin server ");
            buff.append(attribute.getServerId());
            loggerCnx.log(BasicLevel.ERROR, buff.toString(), e);
            throw new OdisJmsException(buff.toString(), e);
        } catch (AdminException e) {
            StringBuffer buff = new StringBuffer("AdminException [");
            buff.append(attribute.getName());
            buff.append("] admin server ");
            buff.append(attribute.getServerId());
            loggerCnx.log(BasicLevel.ERROR, buff.toString(), e);
            throw new OdisJmsException(buff.toString(), e);
        } catch (JMSException e) {
            StringBuffer buff = new StringBuffer("JMSException [");
            buff.append(attribute.getName());
            buff.append("] during admin server connexion");
            buff.append(attribute.getServerId());
            loggerCnx.log(BasicLevel.ERROR, buff.toString(), e);
            throw new OdisJmsException(buff.toString(), e);
        }
    }

    /**
     * delegate to inherited class to implement the destination. It can be a Queue or a Topic.
     * @throws ConnectException the connect exception
     * @throws AdminException the admin exception
     * @throws JMSException the JMS exception
     */
    abstract void specificCreateDestination() throws ConnectException, AdminException, JMSException;

    /**
     * getter.
     * @return int
     */
    public int getState() {
        return state;
    }

    /**
     * setter.
     * @param status int
     */
    public void setState(final int status) {
        this.state = status;
    }

    /**
     * Initialize the component.
     * @throws OdisCnxException if the checkProperties failed
     */
    public void initComponent() throws OdisConnectionException {
        try {
            attribute.checkProperties();
        } catch (OdisAttributeException e) {
            throw new OdisConnectionException(e);
        }
        createDestination();
    }

    /**
     * launch the component.
     * @throws OdisCnxException if a JMSException occured
     */
    public void launchComponent() throws OdisConnectionException {
        try {
            cnx.start();
        } catch (JMSException e) {
            loggerCnx.log(BasicLevel.ERROR, "unable to launchConnexionJMS", e);
            throw new OdisConnectionException("unable to launchConnexionJMS", e);
        }
    }

    /**
     * stop the connection.
     */
    synchronized public void stopComponent() {
        try {
            attribute.removeSession();
            if (loggerCnx.isLoggable(BasicLevel.DEBUG)) {
                loggerCnx.log(BasicLevel.INFO, "Closing JMS session " + session);
            }
            session.close();
            if (loggerCnx.isLoggable(BasicLevel.DEBUG)) {
                loggerCnx.log(BasicLevel.DEBUG, "JMS session closed");
            }
        } catch (JMSException e) {
            loggerCnx.log(BasicLevel.ERROR, "Unable to stop ConnectionJMS", e);
        }
        if (cnx != null) {
            if (attribute.getNbSession() <= 0) {
                ConnectionJMSManager.getInstance().disconnect(this);
            } else {
                if (loggerCnx.isLoggable(BasicLevel.DEBUG)) {
                    loggerCnx.log(BasicLevel.DEBUG, "JMS Connection not closed : nb of remaining sessions = "
                            + attribute.getNbSession());
                }
            }
        }
    }

    /**
     * break all dependency of the other object.
     */
    public void unloadComponent() {
        attribute.deleteObserver(this);
        session = null;
        destination = null;
        cnx = null;
    }

    /**
     * setter.
     * @param state AbstractCnxState
     */
    public void setConnectionState(AbstractConnectionState state) {
        this.cnxState = state;
        if (loggerCnx.isLoggable(BasicLevel.INFO)) {
            StringBuffer buffer = new StringBuffer(toString());
            buffer.append(" set to ");
            buffer.append(state.getStateName());
            loggerCnx.log(BasicLevel.INFO, buffer.toString());
        }
    }

    /**
     * getter.
     * @return AbstractCnxState
     */
    public AbstractConnectionState getConnectionState() {
        return cnxState;
    }

    /**
     * update.
     * @param o Observable
     * @param arg the arg
     * @see java.util.Observer#update(java.util.Observable, java.lang.Object)
     */
    public void update(Observable o, Object arg) {
    }
}
