package org.ow2.odis.joram;

import java.io.Serializable;

import javax.jms.BytesMessage;
import javax.jms.JMSException;
import javax.jms.Message;
import javax.jms.MessageConsumer;
import javax.jms.ObjectMessage;
import javax.jms.TextMessage;

import org.objectweb.util.monolog.Monolog;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.odis.connection.IConnectionIn;
import org.ow2.odis.connection.context.Context;
import org.ow2.odis.connection.context.ContextMessage;
import org.ow2.odis.encoder.GenericEncoder;
import org.ow2.odis.exception.OdisAttributeException;
import org.ow2.odis.port.PortIn;

/**
 * The Class AbstractConnexionJMSIn.
 * @author Olivier Beyler
 */
public abstract class AbstractConnectionJMSIn extends AbstractConnectionJMS implements IConnectionIn {

    /** Logger for this class. */
    static final Logger loggerCnxJMSIn = Monolog.initialize().getLogger(AbstractConnectionJMSIn.class.getName());
    /** The message. */
    private Message msg = null;
    /** The receiver. */
    private MessageConsumer receiver;

    /**
     * Instantiates a new abstract connection JMS in.
     * @param attribute the attribute
     */
    public AbstractConnectionJMSIn(final AbstractJMSConnectionAttribute attribute) {
        super(attribute);
        if (loggerCnxJMSIn.isLoggable(BasicLevel.DEBUG)) {
            loggerCnxJMSIn.log(BasicLevel.DEBUG, "Constructor");
        }
    }
    /** The port in. */
    PortIn portIn = null;

    /**
     * If there is no preloaded message, no connection to the admin server is allowed<br>
     * otherwise gives the number of messages present into the JMS server and the message.
     * @return the number of pending message present in destination. <br>
     */
    public final int getPendingMessage() {
        int result = 0;
        if (msg == null) {
            result = 0;
        } else {
            // add 1 to show the presence of the message
            result = 1;
        }
        return result;
    }

    /**
     * Sets the port in.
     * @param portIn the port in
     * @return the connection in
     * @see com.francetelecom.odis.component.connexion.IConnexionIn#setPortIn(com.francetelecom.odis.component.port.PortIn)
     */
    public abstract IConnectionIn setPortIn(PortIn portIn);

    /**
     * Acknowledges the message.
     * @param valid *** NOW UNUSED ***
     * @see com.francetelecom.odis.component.connexion.IConnexionIn#ackMessage(boolean)
     */
    public final void ackMessage(final boolean valid) {
        MsgListener listener;
        try {
            msg = null;
            listener = (MsgListener) getReceiver().getMessageListener();
            listener.notifySem();
        } catch (JMSException e) {
            loggerCnxJMSIn.log(BasicLevel.ERROR, "JMSException on " + attribute.getName(), e);
        }
    }

    /**
     * Gets the byte message.
     * @return the context message
     * @see com.francetelecom.odis.component.connexion.IConnexionIn#getContextMessage()
     */
    public final ContextMessage getContextMessage() {
        ContextMessage result = null;
        byte[] message = null;
        try {
            if (msg != null) {
                synchronized (this) {
                    
                    if (msg != null) {
                        if (loggerCnxJMSIn.isLoggable(BasicLevel.DEBUG)) {
                            loggerCnxJMSIn.log(BasicLevel.DEBUG, "MessageId (" + msg.getJMSMessageID() + ") on " + attribute.getName());
                        }
                        try{
                        switch (attribute.getModeMessage()) {
                            case AbstractJMSConnectionAttribute.MODE_BYTE_MESSAGE:
                                 {
                                    final int length = (int) ((BytesMessage) msg).getBodyLength();
                                    message = new byte[length];
                                    ((BytesMessage) msg).readBytes(message);
                                }
                                break;
                            case AbstractJMSConnectionAttribute.MODE_TEXT_MESSAGE:
                                 {
                                    final String text = ((TextMessage) msg).getText();
                                    message = attribute.getEncoder().encode(text);
                                }
                                break;
                            case AbstractJMSConnectionAttribute.MODE_OBJECT_MESSAGE: {
                                loggerCnxJMSIn.log(BasicLevel.ERROR, "receive ObjectMessage");
                                final Serializable object = ((ObjectMessage) msg).getObject();                                
                                message=attribute.getEncoder().encode(object);
                                break;

                            }
                        }
                        } catch (ClassCastException e){
                            message = autoConfig();
                        }

                        if (loggerCnxJMSIn.isLoggable(BasicLevel.DEBUG)) {
                            if (message != null) {
                                loggerCnxJMSIn.log(BasicLevel.DEBUG, "receive msg size " + message.length);
                            } else {
                                loggerCnxJMSIn.log(BasicLevel.DEBUG, "receive msg result null (not a ByteMessage or TextMessage)");
                            }
                            loggerCnxJMSIn.log(BasicLevel.DEBUG, "receive on link " + attribute.getName());
                        }
                    } else {
                        if (loggerCnxJMSIn.isLoggable(BasicLevel.DEBUG)) {
                            loggerCnxJMSIn.log(BasicLevel.DEBUG, "receive NULL msg !!! ");
                        }
                    }
                }
            }
        } catch (JMSException e) {
            loggerCnxJMSIn.log(BasicLevel.ERROR, "JMSException ! ", e);        
        } catch (RuntimeException e) {
            loggerCnxJMSIn.log(BasicLevel.ERROR, "RuntimeException ! ", e);
        }

        result = new ContextMessage(message, new Context(attribute.getName(), ""));
        return result;
    }
    
    
    private byte[] autoConfig() throws JMSException{
        loggerCnxJMSIn.log(BasicLevel.ERROR, "The declare config for JMS connexion "+attribute.getName()+" is bad !");
        loggerCnxJMSIn.log(BasicLevel.WARN, "USE AUTOCONFIG INSTEAD OF THE DECLARE CONFIG !!! ");
        loggerCnxJMSIn.log(BasicLevel.WARN, "MODE_MESSAGE was:"+attribute.getStrModeMessage());
        byte[] message= null;
        try {
        if (msg instanceof BytesMessage) {
                attribute.setModeMessage(AbstractJMSConnectionAttribute.PROPERTY_MODE_BYTE_MESSAGE);

              final int length = (int) ((BytesMessage) msg).getBodyLength();
              message = new byte[length];
              ((BytesMessage) msg).readBytes(message);
        } else if (msg instanceof TextMessage) {
             attribute.setModeMessage(AbstractJMSConnectionAttribute.PROPERTY_MODE_TEXT_MESSAGE);
             final String text = ((TextMessage) msg).getText();
             message = new GenericEncoder().encode(text);
        } else if (msg instanceof ObjectMessage) {
            attribute.setModeMessage(AbstractJMSConnectionAttribute.PROPERTY_MODE_OBJECT_MESSAGE);            
            final Serializable object = ((ObjectMessage) msg).getObject();
            message=new GenericEncoder().encode(object);                                
            }        
        } catch (OdisAttributeException ex) {
                ex.printStackTrace();
            }
        loggerCnxJMSIn.log(BasicLevel.WARN, "MODE_MESSAGE is forced by autoconfig to :"+attribute.getStrModeMessage());
        return message;
    }

    /**
     * getter.
     * @return message Message
     */
    public final Message getMsg() {
        return msg;
    }

    /**
     * Allows the listener to set the last message received.
     * @param msg last message receive by the message Listener
     * @see com.francetelecom.odis.component.connexion.jms.MsgListener
     */
    public final void setMsg(final Message msg) {
        this.msg = msg;
    }

    /**
     * getter.
     * @return MessageConsumer
     */
    public final MessageConsumer getReceiver() {
        return receiver;
    }

    /**
     * setter.
     * @param receiver MessageConsumer
     */
    public final void setReceiver(final MessageConsumer receiver) {
        this.receiver = receiver;
    }

    /**
     * Stop component.
     * @see com.francetelecom.odis.component.connexion.jms.AbstractConnectionJMS#stopComponent()
     */
    public final void stopComponent() {
        MsgListener listener;
        try {
            if (loggerCnxJMSIn.isLoggable(BasicLevel.INFO)) {
                loggerCnxJMSIn.log(BasicLevel.INFO, "Close ConnexionJMSIn requiered " + attribute.getName());
            }
            listener = (MsgListener) getReceiver().getMessageListener();
            if (listener != null) {
                // No need to close the receiver according to the javax.jms.Session.close() contract
                // getReceiver().close();
                listener.setAlive(false);
                listener.notifySem();
            }
            if (loggerCnxJMSIn.isLoggable(BasicLevel.INFO)) {
                loggerCnxJMSIn.log(BasicLevel.INFO, "Close ConnexionJMSIn done " + attribute.getName());
            }
        // No need for that either
        // getReceiver().setMessageListener(null);
        } catch (JMSException e) {
            loggerCnxJMSIn.log(BasicLevel.ERROR, "JMSException on " + attribute.getName(), e);
        }
        super.stopComponent();
    }
}
