package org.ow2.odis.joram;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import javax.jms.Connection;
import javax.jms.ConnectionFactory;
import javax.jms.ExceptionListener;
import javax.jms.JMSException;

import org.objectweb.joram.client.jms.ha.tcp.HATcpConnectionFactory;
import org.objectweb.joram.client.jms.local.LocalConnectionFactory;
import org.objectweb.joram.client.jms.tcp.TcpConnectionFactory;
import org.objectweb.util.monolog.Monolog;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.odis.admin.Launch;
import org.ow2.odis.exception.OdisJmsException;


/**
 * The Class ConnexionJMSManager.
 */
public class ConnectionJMSManager {
    /** The instance. */
    static private ConnectionJMSManager instance = null;

    /** The Constant LOGGER. */
    final static Logger LOGGER = Monolog.initialize().getLogger(ConnectionJMSManager.class.getName());

    /**
     * map between a string id and a connection factory. This id is created by using connection attribute.
     */
    private final transient Map mapCnxf;

    /**
     * map between a string id and a connection. This id is created by using connection attribute.
     */
    private final transient Map mapCnx;

    /**
     * map between a connection and a set of destination. It's useful to inform all destination concern by this connection that the
     * connection failed, a new connection will replace it.
     */
    private final transient Map mapCnxDestination;

    /**
     * implement Singleton pattern.
     * @return unique instance of Connection Manager
     */
    public static ConnectionJMSManager getInstance() {
        if (instance == null) {
            synchronized (LOGGER) {
                if (instance == null) {
                    instance = new ConnectionJMSManager();
                }
            }
        }
        return instance;
    }

    /**
     * Constructor.
     */
    private ConnectionJMSManager() {
        super();
        mapCnxf = new HashMap();
        mapCnx = new HashMap();
        mapCnxDestination = new HashMap();
    }

    /**
     * Creates a new Connection by using the connexionJMS' attribute description. If a High availability cluster is defined the cluster
     * connection factory is used.
     * @param connectionJMS the JMS connection
     * @return Connection
     */
    public synchronized Connection connect(final AbstractConnectionJMS connectionJMS) {
        Connection cnx = null;
        final AbstractJMSConnectionAttribute attribute = (AbstractJMSConnectionAttribute) connectionJMS.getAttribute();
        // look for connection factory
        // et si on ne la trouve pas on en cree une.
        String keyCnxFactory;
        final boolean isClustering = !("".equals(attribute.getHAClusterDescription()));
        ConnectionFactory cnxf = null;
        if (isClustering) {
            keyCnxFactory = attribute.getHAClusterDescription();
            cnxf = (ConnectionFactory) mapCnxf.get(keyCnxFactory);
            if (cnxf == null) {
                if (LOGGER.isLoggable(BasicLevel.INFO)) {
                    LOGGER.log(BasicLevel.INFO, "Creating clustered ConnexionFactory");
                }
                cnxf = HATcpConnectionFactory.create(attribute.getHAClusterDescription());
                mapCnxf.put(keyCnxFactory, cnxf);
            }
        } else {
            keyCnxFactory = attribute.getProxyServerHostName() + "@" + attribute.getProxyServerPort();
            cnxf = (ConnectionFactory) mapCnxf.get(keyCnxFactory);
            if (cnxf == null) {
                if (LOGGER.isLoggable(BasicLevel.INFO)) {
                    LOGGER.log(BasicLevel.INFO, "Creating ConnexionFactory ("
                            + (Launch.joramCollocated ? "collocated)" : "NON collocated)"));
                }
                if (Launch.joramCollocated) {
                    cnxf = LocalConnectionFactory.create();
                } else {
                    cnxf = TcpConnectionFactory.create(attribute.getProxyServerHostName(), attribute.getProxyServerPort());
                }
                mapCnxf.put(keyCnxFactory, cnxf);
            }
        }
        if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
            LOGGER.log(BasicLevel.DEBUG, "ConnexionFactory is " + cnxf.hashCode());
        }
        final String keyConnexion = attribute.getName() + "@" + keyCnxFactory + "@" + attribute.getUserName() + "@"
                + attribute.getUserPassword();
        try {
            cnx = (Connection) mapCnx.get(keyConnexion);
            if (cnx == null) {
                if (LOGGER.isLoggable(BasicLevel.INFO)) {
                    LOGGER.log(BasicLevel.INFO, "Creating connection for " + keyConnexion);
                }
                cnx = cnxf.createConnection(attribute.getUserName(), attribute.getUserPassword());
                mapCnx.put(keyConnexion, cnx);
                if (LOGGER.isLoggable(BasicLevel.INFO)) {
                    LOGGER.log(BasicLevel.INFO, "Connection created :" + cnx);
                }
                final ExceptionListener liste = cnx.getExceptionListener();
                cnx.setExceptionListener(new JMSExceptionListener(cnx, liste));
            }
            Set setDestination = (Set) mapCnxDestination.get(cnx);
            if (setDestination == null) {
                setDestination = new HashSet();
                if (LOGGER.isLoggable(BasicLevel.INFO)) {
                    LOGGER.log(BasicLevel.INFO, "List of destinations created:" + setDestination);
                }
                mapCnxDestination.put(cnx, setDestination);
            }
            setDestination.add(connectionJMS);
            // ------------------------------------------------------------
            // pour debug permet de donner plus d infos dans les traces
            if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                LOGGER.log(BasicLevel.DEBUG, "add " + connectionJMS + " to the list of destination ");
                final Iterator itDestination = setDestination.iterator();
                AbstractConnectionJMS cnxJms;
                while (itDestination.hasNext()) {
                    cnxJms = (AbstractConnectionJMS) itDestination.next();
                    LOGGER.log(BasicLevel.DEBUG, " --->  " + cnxJms.attribute.getName() + " element of the list of destination ");
                }
            }
            // fin de pour debug
            // ------------------------------------------------------------
        } catch (JMSException e) {
            LOGGER.log(BasicLevel.FATAL, "Unable to create connection", e);
            cnx = null;
        }
        return cnx;
    }

    /**
     * Closes a JMS connection.
     * @param connexionJMS the ODIS ConnectionJMS using that connection
     */
    public synchronized void disconnect(AbstractConnectionJMS connexionJMS) {
        Connection cnx = null;
        final AbstractJMSConnectionAttribute attribute = (AbstractJMSConnectionAttribute) connexionJMS.getAttribute();
        String keyCnxFactory;
        final boolean isClustering = !("".equals(attribute.getHAClusterDescription()));
        if (isClustering) {
            keyCnxFactory = attribute.getHAClusterDescription();
        } else {
            keyCnxFactory = attribute.getProxyServerHostName() + "@" + attribute.getProxyServerPort();
        }
        final String keyConnexion = attribute.getName() + "@" + keyCnxFactory + "@" + attribute.getUserName() + "@" + attribute.getUserPassword();
        cnx = (Connection) mapCnx.get(keyConnexion);
        // Always prune out of our maps
        mapCnx.remove(keyConnexion);
        mapCnxDestination.remove(cnx);
        // Then close for Joram
        if (cnx == null) {
            if (LOGGER.isLoggable(BasicLevel.WARN)) {
                LOGGER.log(BasicLevel.WARN, "Trying to close non mapped connection : " + keyConnexion);
            }
        } else {
            try {
                cnx.setExceptionListener(null);
                cnx.close();
                if (LOGGER.isLoggable(BasicLevel.INFO)) {
                    LOGGER.log(BasicLevel.INFO, "JMS connection closed : " + keyConnexion);
                }
            } catch (JMSException e) {
                LOGGER.log(BasicLevel.ERROR, "JMSException while closing connection : " + keyConnexion + " - This should be recoverable.", e);
            }              
        }
    }

    /**
     * Tries to create a new connection and inform all links involved with the oldConnexion to use this new connection.
     * @param oldConnexion the old connection
     * @return -1 if failed 0 if succeed
     */
    public int restoreConnexion(final Connection oldConnexion) {
        if (LOGGER.isLoggable(BasicLevel.INFO)) {
            LOGGER.log(BasicLevel.INFO, "Trying to restore connection : " + oldConnexion + " ie " + oldConnexion.hashCode());
        }
        int result = 0;
        final Set setDestination = (Set) mapCnxDestination.get(oldConnexion);
        if (setDestination != null) {
            Connection newConnexion = null;
            Iterator itDestination = setDestination.iterator();
            AbstractConnectionJMS cnx;
            if (setDestination.isEmpty()) {
                if (LOGGER.isLoggable(BasicLevel.WARN)) {
                    LOGGER.log(BasicLevel.WARN, "Empty destination set");
                }
                result = -1;
            }
            while (itDestination.hasNext() && result == 0) {
                cnx = (AbstractConnectionJMS) itDestination.next();
                newConnexion = connect(cnx);
                if (newConnexion == null) {
                    if (LOGGER.isLoggable(BasicLevel.ERROR)) {
                        LOGGER.log(BasicLevel.ERROR, "Restore connection failed for " + oldConnexion);
                    }
                    result = -1;
                }
                if (LOGGER.isLoggable(BasicLevel.INFO)) {
                    if (newConnexion != null) {
                        LOGGER.log(BasicLevel.INFO, "Connection restored " + newConnexion + " ie " + newConnexion.hashCode());
                    }
                }
            }
            // affecte la nouvelle connexion cree a tous les liens et recree les
            // destinations
            itDestination = setDestination.iterator();
            while (itDestination.hasNext() && result == 0) {
                cnx = (AbstractConnectionJMS) itDestination.next();
                cnx.cnx = newConnexion;
                try {
                    cnx.createDestination();
                } catch (OdisJmsException e) {
                    // TODO a tester!!!!
                    LOGGER.log(BasicLevel.ERROR, "Create destination failed", e);
                    result = -1;
                }
            }
            if (result == 0) {
                mapCnxDestination.remove(oldConnexion);
                mapCnxDestination.put(newConnexion, setDestination);
                // active connection
                itDestination = setDestination.iterator();
                try {
                    newConnexion.start();
                    if (LOGGER.isLoggable(BasicLevel.INFO)) {
                        LOGGER.log(BasicLevel.INFO, "Connection started : " + newConnexion);
                    }
                } catch (JMSException e) {
                    LOGGER.log(BasicLevel.ERROR, "Connection start failed for " + newConnexion, e);
                    result = -1;
                }
            }
            if (result == 0) {
                // informe le lien que la connexion est reouverte
                itDestination = setDestination.iterator();
                AbstractJMSConnectionAttribute attribute = null;
                while (itDestination.hasNext()) {
                    cnx = (AbstractConnectionJMS) itDestination.next();
                    cnx.setState(AbstractConnectionJMS.READY);
                    if (LOGGER.isLoggable(BasicLevel.INFO)) {
                        LOGGER.log(BasicLevel.INFO, "Connection restored : " + cnx.getAttribute().getName());
                    }
                    attribute = (AbstractJMSConnectionAttribute) cnx.getAttribute();
                    attribute.incNbrReconnexion(1);
                }
            }
        }
        return result;
    }

    /**
     * reinitialize all connexionJMS witch use this connection.
     * @param object the object
     * @return -1 if failed 0 if succeed
     */
    public int suspendCnx(final Object object) {
        int result = 0;
        // reinitialize all link using this connexion
        final Set setDestination = (Set) mapCnxDestination.get(object);
        if (setDestination == null) {
            LOGGER.log(BasicLevel.FATAL, "try to suspend unknown connection");
            result = -1;
        }
        if (result == 0) {
            final Iterator itDestination = setDestination.iterator();
            AbstractConnectionJMS cnx = null;
            while (itDestination.hasNext()) {
                cnx = (AbstractConnectionJMS) itDestination.next();
                if (cnx.attribute != null) {
                    if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                        LOGGER.log(BasicLevel.DEBUG, "inform " + cnx.attribute.getName() + " that the connexion is closed");
                    }
                }
                cnx.setState(AbstractConnectionJMS.NOT_READY);
                cnx.session = null;
                cnx.cnx = null;
                cnx.destination = null;
            }
            // remove connection from the map of connection
            final Set keys = mapCnx.keySet();
            final Object[] array = keys.toArray();
            Object connexion = null;
            for (int i = 0; i < array.length; i++) {
                if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                    LOGGER.log(BasicLevel.DEBUG, "Connexion " + array[i]);
                }
                connexion = mapCnx.get(array[i].toString());
                if (connexion == null) {
                    LOGGER.log(BasicLevel.FATAL, "Connection unknown !!!");
                }
                if (connexion.equals(object)) {
                    mapCnx.remove(array[i]);
                }
            }
        }
        return result;
    }

    /**
     * Active connection.
     */
    public void activeConnection() {
        final Set keys = mapCnx.keySet();
        final Object[] array = keys.toArray();
        Connection connexion = null;
        for (int i = 0; i < array.length; i++) {
            if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                LOGGER.log(BasicLevel.DEBUG, "Connection " + array[i] + " is starting");
            }
            connexion = (Connection) mapCnx.get(array[i].toString());
            try {
                connexion.start();
                if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                    LOGGER.log(BasicLevel.DEBUG, "Connection " + array[i] + " start");
                }
            } catch (JMSException e) {
                LOGGER.log(BasicLevel.ERROR, "JMS EXCEPTION", e);
            }
        }
    }
}
