package org.ow2.odis.joram;

import javax.jms.JMSException;
import javax.jms.Message;
import javax.jms.MessageListener;

import org.objectweb.util.monolog.Monolog;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.odis.port.PortIn;
import org.ow2.odis.util.Semaphore;

/**
 * The Class MsgListener.
 * @author Beyler Olivier FranceTelecom R&D
 */
public class MsgListener implements MessageListener {
    /** Logger for this class. */
    static final Logger LOGGER = Monolog.initialize().getLogger(MsgListener.class.getName());

    /** Keeps a reference to an AbstractConnexionJMSIn. */
    private AbstractConnectionJMSIn destination = null;

    /** Denotes that the listener is allowed to handle messages. */
    private boolean isAlive = true;

    /** Used to block the message listener. */
    private Semaphore sem = new Semaphore();

    /**
     * Instantiates a new message listener.
     * @param destination the destination
     */
    MsgListener(AbstractConnectionJMSIn destination) {
        super();
        if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
            LOGGER.log(BasicLevel.DEBUG, "create messageListener  " + this);
        }
        this.destination = destination;
    }

    /**
     * Allows MsgListener to acknowledge current message and get next one.
     */
    public void notifySem() {
        sem.release();
        if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
            LOGGER.log(BasicLevel.DEBUG, "notify MessageListener");
        }
    }

    /**
     * Method called by the JMS provider when a new message appears on destination.
     * @param pArg0 Message
     * @see javax.jms.MessageListener#onMessage(javax.jms.Message)
     */
    public void onMessage(Message pArg0) {
        
        setMsg(pArg0);
        try {
            if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                LOGGER.log(BasicLevel.DEBUG, "receive Msg with Id: " + pArg0.getJMSMessageID() + " ref:" + pArg0);
            }
        } catch (JMSException e2) {
            LOGGER.log(BasicLevel.ERROR, "onMessage JMSExeption occured during getJMSMessageId", e2);
        }
        try {
            while (getPort() == null && isAlive) {
                Thread.sleep(1000);
                if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                    LOGGER.log(BasicLevel.DEBUG, "waiting for Link...");
                }
            }
            if (isAlive){
                if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                    LOGGER.log(BasicLevel.DEBUG, "informing Link " + getPort());
                }
                getPort().onMessage();
                if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                    LOGGER.log(BasicLevel.DEBUG, "going on wait on link "+destination.attribute.getName());
                }
                sem.waiting();
                if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                    LOGGER.log(BasicLevel.DEBUG, "sem released on link "+destination.attribute.getName());
                }
            } else {
                if (LOGGER.isLoggable(BasicLevel.INFO)) {
                    LOGGER.log(BasicLevel.INFO, "NOT ALIVE - ignoring message on link "+destination.attribute.getName());
                }
            }
        } catch (InterruptedException e1) {
            LOGGER.log(BasicLevel.ERROR, "unable to wait semaphore", e1);
        }
        if (isAlive) {
            try {
                if (pArg0 != null) {
                    if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                        LOGGER.log(BasicLevel.DEBUG, "acknowledge  message with id :" + pArg0.getJMSMessageID() + " ref:" + pArg0);
                    }
                    pArg0.acknowledge();
                } else {
                    LOGGER.log(BasicLevel.ERROR, "can't acknowledge null message");
                }
            } catch (JMSException e) {
                LOGGER.log(BasicLevel.ERROR, "onMessage JMSExeption occured during acknowledgment", e);
            }
        } else {
            if (LOGGER.isLoggable(BasicLevel.INFO)) {
                LOGGER.log(BasicLevel.INFO, "NOT ALIVE - won't acknowledge message on link "+destination.attribute.getName());
            }
        } 
    }

    /**
     * Gets the port.
     * @return the port
     */
    private PortIn getPort() {
        return destination.portIn;
    }

    /**
     * Affects the Message only if a destination exist.
     * @param obj the message
     */
    private void setMsg(Message obj) {
        if (destination != null) {
            destination.setMsg((Message) obj);
        }
    }

    /**
     * Sets the Message Listener live state, allowing to handle and acknowledge messages.
     * @param pAlive the new state
     */
    public void setAlive(boolean pAlive) {
        this.isAlive = pAlive;
    }
}
